import unittest
from classifier.constructor import Hashtron

class TestHashtronForward(unittest.TestCase):
    def test_forward(self):
        # Create a Hashtron instance with a sample program
        tron = Hashtron.new([[1, 4], [3, 1]], 2)

        # Call the bytes_buffer method
        result = tron.forward.forward(0, False)

        # Ensure the output is not None
        self.assertIsNotNone(result, "forward returned None")

        # Check if the expected integer is the output
        self.assertIs(0, result)


if __name__ == '__main__':
    unittest.main()



from classifier.constructor import Hashtron
from layer.layer import Layer
from hash.hash import Hash
from net.single_value import SingleValue
from net.input import Input

class FeedforwardNetwork:
    def __init__(self, net):
        self.net = net
        self.layers = []
        self.mapping = []
        self.combiners = []
        self.premodulo = []

    def new_layer(self, n: int, bits: int, premodulo: int = 0) -> None:
        layer = [Hashtron.new(None, bits) for _ in range(n)]
        self.layers.append(layer)
        self.mapping.append(bits)
        self.combiners.append(None)
        self.premodulo.append(premodulo)

    def new_combiner(self, layer: Layer) -> None:
        self.layers.append([])
        self.mapping.append(0)
        self.combiners.append(layer)
        self.premodulo.append(0)


    def len_layers(self) -> int:
        """
        Get the number of layers in the network.

        :return: The number of layers.
        """
        return len(self.layers)

    def infer(self, in_val) -> SingleValue:
        """
        Infer the network output based on input.

        :param in_val: The input to the network.
        :return: The output of the network.
        """
        if (not hasattr(in_val, 'feature') or not callable(in_val.feature)) and in_val is not Input and in_val is not SingleValue:
            in_val = SingleValue(in_val)
        if (not hasattr(in_val, 'parity') or not callable(in_val.parity)) and in_val is not Input:
            in_val = Input(in_val)
        output = in_val
        for l_prev in range(0, self.len_layers(), 2):
            output, _ = self.forward(output, l_prev, -1, 0)
        return output

    def forward(self, in_val, l: int, worst: int, neg: int) -> (Input, bool):
        """
        Forward pass through the network.

        :param in_val: The input to the layer.
        :param l: The layer index.
        :param worst: The index of the worst hashtron.
        :param neg: Whether to negate the worst hashtron's output.
        :return: The intermediate output and a boolean indicating if the worst hashtron's output was computed.
        """
        if len(self.combiners) > l + 1 and self.combiners[l + 1] is not None:
            combiner = self.combiners[l + 1].lay()
            computed = False

            for i in range(len(self.layers[l])):
                feat = in_val.feature(i)
                if self.premodulo[l] != 0:
                    feat = Hash.hash(feat, i, self.premodulo[l])
                bit = self.layers[l][i].forward.forward(feat, (i == worst) and (neg == 1))
                combiner.put(i, (bit & 1) != 0)
                if i == worst:
                    computed = (bit & 1) != 0

            return combiner, computed
        elif len(self.mapping) > l and self.mapping[l] > 0:
            if self.premodulo[l] != 0:
                in_val = SingleValue(Hash.hash(in_val.feature(0), 0, self.premodulo[l]))
            val = self.layers[l][0].forward.forward(in_val.feature(0), (0 == worst) and (neg == 1))
            return SingleValue(val), False
        else:
            if self.premodulo[l] != 0:
                in_val = SingleValue(Hash.hash(in_val.feature(0), 0, self.premodulo[l]))
            bit = self.layers[l][0].forward.forward(in_val.feature(0), (0 == worst) and (neg == 1))
            return SingleValue(bit & 1), (bit & 1) != 0





import unittest
import os
from net.feedforward.net import Net
from layer.majpool2d.layer import MajPool2DLayer
from datasets.mnist.mnist import MNISTDataset
import urllib.request
import tempfile

class TestNetConstruct(unittest.TestCase):
    def test_construct(self):
        # Specify network size
        fanout1 = 3
        fanout2 = 5
        fanout3 = 3
        fanout4 = 5
        # Create a Hashtron network (MNIST handwritten digits net)
        tron = Net.new()
        tron.new_layer(fanout1*fanout2*fanout3*fanout4, 0, 1<<fanout4)
        tron.new_combiner(MajPool2DLayer(fanout1*fanout2*fanout4, 1, fanout3, 1, fanout4, 1, 1))
        tron.new_layer(fanout1*fanout2, 0, 1<<fanout2)
        tron.new_combiner(MajPool2DLayer(fanout2, 1, fanout1, 1, fanout2, 1, 1))
        tron.new_layer(1, 0)
        # load weights from zlib file
        filename = os.path.expanduser('~/classifier/cmd/train_mnist/output.77.json.t.zlib')
        if os.path.exists(filename):
            ok = tron.io.read_zlib_weights_from_file(filename)
        else:
            # load online model weights zlib file
            with urllib.request.urlopen('https://www.hashtron.cloud/dl/mnist/output.77.json.t.zlib') as f:
                with tempfile.NamedTemporaryFile(delete=True) as dst_file:
                    data = f.read()
                    dst_file.write(data)
                    ok = tron.io.read_zlib_weights_from_file(dst_file.name)
        
        if not ok:
            raise Exception('model weights were not loaded')
        
        # test the datasets
        for i in range(2):
            # load offline then online mnist
            try:
                dataset = MNISTDataset(True, i == 0)
            except FileNotFoundError:
                dataset = MNISTDataset(True, i == 0, MNISTDataset.download())
            
            correct = 0
            for sample in dataset[0:500]:
                pred = tron.network.infer(sample).feature(0) & 1
                actual = sample.output().feature(0) & 1
                if pred == actual:
                    correct+=1
            print(100 * correct // len(dataset), '% on 500', ['train', 'eval'][i], 'MNIST samples')

if __name__ == '__main__':
    unittest.main()



import unittest
from hash.hash import Hash

class TestHash(unittest.TestCase):
    def test_hash(self):
        # Test cases for hash function
        self.assertEqual(Hash.hash(0, 0, 0), 0)
        self.assertLess(Hash.hash(1, 1, 10), 10)

    def test_string_hash(self):
        # Test cases for string_hash function
        self.assertEqual(Hash.string_hash(0, "test"), Hash.string_hash(0, "test"))

    def test_loop_length(self):
        # Loop length test
        bound1 = 10
        bound2 = 100000
        count = 0

        max_val = 1
        while max_val <= (1 << bound1):
            visited = [False] * max_val
            current = 0

            for s in range(bound2):
                current = Hash.hash(current, s, max_val)
                if current == 0 or visited[current]:
                    visited = [False] * max_val
                    continue
                else:
                    visited[current] = True
                    count += 1

            max_val <<= 1

        print(f"Tested bound1 1 << {bound1}, bound2 {bound2}, result: {count} (higher is likely better)")

if __name__ == '__main__':
    unittest.main()



import random
from classifier.model import HashtronModel
from classifier.forward import HashtronForward
from classifier.view import HashtronView

class Hashtron:
    def __init__(self, program=None, bits=1):
        self.program = program if program else [[random.randint(0, 0xFFFFFFFF) >> 1, 2]]
        self.bits = bits if bits > 0 else 1
        self.model = HashtronModel(self)
        self.forward = HashtronForward(self)
        self.view = HashtronView(self)
    
    @staticmethod
    def new(program=None, bits=1):
        return Hashtron(program, bits)




class MajPool2D:
    def __init__(self, vec, width, height, subwidth, subheight, capwidth, capheight, repeat, bias):
        """
        Initialize the MajPool2D combiner.

        :param vec: The vector of boolean values.
        :param width: The width of the pooling matrix.
        :param height: The height of the pooling matrix.
        :param subwidth: The width of the submatrix.
        :param subheight: The height of the submatrix.
        :param capwidth: The width of the capture area.
        :param capheight: The height of the capture area.
        :param repeat: The number of repetitions.
        :param bias: The bias value for majority pooling.
        """
        self.vec = vec
        self.width = width
        self.height = height
        self.subwidth = subwidth
        self.subheight = subheight
        self.capwidth = capwidth
        self.capheight = capheight
        self.repeat = repeat
        self.bias = bias

    def put(self, n: int, v: bool) -> None:
        """
        Set the n-th boolean value in the vector.

        :param n: The index of the boolean value to set.
        :param v: The boolean value to set.
        """
        self.vec[n] = v

    def disregard(self, n: int) -> bool:
        """
        Check if setting the n-th boolean value to False would not affect any feature output.

        :param n: The index of the boolean value to check.
        :return: True if setting the value to False would not affect the output, False otherwise.
        """
        orign = n
        submatrix = self.subwidth * self.subheight
        matrix = self.width * self.height * submatrix
        base = (n // matrix) * matrix
        n %= matrix
        n //= submatrix
        n *= submatrix
        w0, w1 = 0, 0

        for m in range(submatrix):
            if orign == base + m + n:
                w0 += 1
                w1 -= 1
                continue
            if self.vec[base + m + n]:
                w0 += 1
                w1 += 1
            else:
                w0 -= 1
                w1 -= 1

        cond1 = (w0 > self.bias) == (w1 > self.bias)
        return cond1

    def feature(self, m: int) -> int:
        """
        Compute the m-th feature from the combiner.

        :param m: The index of the feature to compute.
        :return: The computed feature as an integer.
        """
        supermatrix = self.width * self.height
        submatrix = self.subwidth * self.subheight
        matrix = supermatrix * submatrix
        base = (m // matrix) * matrix
        starty = m // self.width
        startx = m % self.width
        o = 0

        for y in range(self.capheight):
            for x in range(self.capwidth):
                xx = (x + startx) % self.width
                yy = (y + starty) % self.height
                w = 0

                for m in range(submatrix):
                    if self.vec[base + submatrix * (self.width * yy + xx) + m]:
                        w += 1
                    else:
                        w -= 1

                o <<= 1
                if w > self.bias:
                    o |= 1

        return o



import unittest
from layer.majpool2d.combiner import MajPool2D

class TestMajPool2D(unittest.TestCase):
    def test_put(self):
        vec = [False] * 400
        combiner = MajPool2D(vec, 10, 10, 2, 2, 5, 5, 1, 0)
        combiner.put(0, True)
        self.assertTrue(vec[0])

    def test_disregard(self):
        vec = [False] * 400
        combiner = MajPool2D(vec, 10, 10, 2, 2, 5, 5, 1, 0)
        self.assertTrue(combiner.disregard(0))

    def test_feature(self):
        vec = [True] * 400
        combiner = MajPool2D(vec, 10, 10, 2, 2, 5, 5, 1, 0)
        self.assertEqual(combiner.feature(0), 33554431)  # Example expected output

if __name__ == '__main__':
    unittest.main()



from net.feedforward.feedforward_network import FeedforwardNetwork
from net.feedforward.io import FeedforwardNetworkIO
from layer.layer import Layer

class Net:
    def __init__(self):
        self.network = FeedforwardNetwork(self)
        self.io = FeedforwardNetworkIO(self)
    
    @staticmethod
    def new():
        return Net()

    def new_layer(self, n: int, bits: int, premodulo: int = 0) -> None:
        self.network.new_layer(n, bits, premodulo)

    def new_combiner(self, layer: Layer) -> None:
        self.network.new_combiner(layer)



class Hash:
    @staticmethod
    def hash(n: int, s: int, max_val: int) -> int:
        # Mixing stage, mix input with salt using subtraction
        m = (n - s) & 0xFFFFFFFF
        # Hashing stage, use xor shift with prime coefficients
        m ^= (m << 2) & 0xFFFFFFFF
        m ^= (m << 3) & 0xFFFFFFFF
        m ^= (m >> 5) & 0xFFFFFFFF
        m ^= (m >> 7) & 0xFFFFFFFF
        m ^= (m << 11) & 0xFFFFFFFF
        m ^= (m << 13) & 0xFFFFFFFF
        m ^= (m >> 17) & 0xFFFFFFFF
        m ^= (m << 19) & 0xFFFFFFFF
        # Mixing stage 2, mix input with salt using addition
        m += s
        m &= 0xFFFFFFFF
        # Modular stage using Lemire's fast alternative to modulo reduction
        return ((m * max_val) >> 32) & 0xFFFFFFFF

    @staticmethod
    def strings_hash(in_val: int, strs: list[str]) -> int:
        out = in_val
        for s in strs:
            out = Hash.string_hash(out, s)
        return out

    @staticmethod
    def string_hash(in_val: int, s: str) -> int:
        out = in_val
        for c in s:
            out = Hash.hash(out, ord(c), 0xFFFFFFFF)
        return out



from layer.majpool2d.combiner import MajPool2D
from layer.layer import Layer

class MajPool2DLayer(Layer):
    def __init__(self, width, height, subwidth, subheight, capwidth, capheight, repeat, bias=0):
        self.width = width
        self.height = height
        self.subwidth = subwidth
        self.subheight = subheight
        self.capwidth = capwidth
        self.capheight = capheight
        self.repeat = repeat
        self.bias = bias

    def lay(self) -> MajPool2D:
        vec = [False] * (self.width * self.height * self.subwidth * self.subheight * self.repeat)
        return MajPool2D(vec, self.width, self.height, self.subwidth, self.subheight, self.capwidth, self.capheight, self.repeat, self.bias)



class Input:
    def __init__(self, obj):
        # Check if the wrapped object has a 'feature' method
        if not hasattr(obj, 'feature') or not callable(obj.feature):
            raise ValueError("The wrapped object must have a 'feature' method.")
        self.obj = obj

    def feature(self, n: int) -> int:
        # Call the wrapped object's feature method
        result = self.obj.feature(n)
        
        # Check if the wrapped object has a 'parity' method
        if hasattr(self.obj, 'parity') and callable(self.obj.parity):
            # If it does, XOR the result with the parity
            result ^= self.obj.parity()
        
        return result

    def parity(self) -> int:
        # Check if the wrapped object has a 'parity' method
        if hasattr(self.obj, 'parity') and callable(self.obj.parity):
            # If it does, return the parity
            return self.obj.parity()
        return 0



import unittest
from classifier.constructor import Hashtron

class TestHashtronSerialize(unittest.TestCase):
    def test_serialize(self):
        # Create a Hashtron instance with a sample program
        tron = Hashtron.new([[1, 2], [3, 4]], 2)
        
        a = tron.view.write_json()

        tron.view.read_json(a)
        
        b = tron.view.write_json()
        
        self.assertEqual(a, b)

if __name__ == '__main__':
    unittest.main()



class HashtronModel:
    def __init__(self, program=None, bits=1):
        self.program = program if program else []
        self.bits = bits

    def get(self, n: int) -> tuple[int, int]:
        return self.program[n][0], self.program[n][1]

    def len(self) -> int:
        return len(self.program)

    def bits(self) -> int:
        return self.bits

    def set_bits(self, bits: int):
        self.bits = bits



from setuptools import setup, find_packages

setup(
    name="hashtron",
    version="0.0.7",
    packages=find_packages(where="hashtron"),
    package_dir={"": "hashtron"},
    install_requires=[
        "requests==2.32.3",
    ],
    author="Neurlang Project",
    author_email="77860779+neurlang@users.noreply.github.com",
    description="A Python inference-only engine for the hashtron binary classifier",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/neurlang/pyclassifier",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import zlib
import json

class FeedforwardNetworkIO:
    def __init__(self, net):
        self.net = net

    def read_zlib_weights_from_file(self, file_name: str) -> bool:
        with open(file_name,'rb') as f:
            file_content=zlib.decompress(f.read())
        if len(file_content) <= 2:
            # empty model or uknown format
            return True
        if file_content[0] != ord('['):
            raise ValueError("unknown model format: please fix your model, or update pyclassifier (pip is hashtron) to recognize this model")
        decoded = file_content.decode('ascii')
        parsed = json.loads(decoded)
        i = 0
        for layer in self.net.network.layers:
            for cell in layer:
                cell.view.read_json(json.dumps(parsed[i]))
                i += 1
        return i == len(parsed)



import json

class HashtronView:
    def __init__(self, hashtron):
        self.hashtron = hashtron
    
    def write_json(self) -> str:
        return json.dumps(self.hashtron.program)
    
    def read_json(self, b) -> None:
        self.hashtron.program = json.loads(b)



from hash.hash import Hash

class HashtronForward:
    def __init__(self, hashtron):
        self.hashtron = hashtron

    def forward(self, sample: int, negate: bool) -> int:
        if not self.hashtron.program:
            return 0
        out = 0
        for j in range(self.hashtron.bits):
            input_val = sample if self.hashtron.bits <= 1 else sample | (j << 16)
            ss, maxx = self.hashtron.program[0]
            input_val = Hash.hash(input_val, ss, maxx)
            for i in range(1, len(self.hashtron.program)):
                s, max_val = self.hashtron.program[i]
                ss ^= s
                maxx -= max_val
                input_val = Hash.hash(input_val, ss, maxx)
            input_val &= 1
            if negate:
                input_val ^= 1
            if input_val != 0:
                out |= 1 << j
        return out
